pentaho = typeof pentaho == "undefined" ? {} : pentaho;

pentaho.viz = typeof pentaho.viz == "undefined" ? {} : pentaho.viz;

pentaho.visualizations = pentaho.visualizations || [];

var chart;
var list;
var formatNumber;
var all;
var dimensions = [];
var dataTable;
var crossFilterWrapper;

/*
    Visualization Metadata
    These objects describe the visualizations provided by this library.
*/
pentaho.visualizations.push({
    id: 'CrossFilter',                          // unique identifier
    type: 'histograms',                       // generic type id
    source: 'crossfilter',                          // id of the source library
    name: 'CrossFilter',                     // visible name, this will come from a properties file eventually
    'class': 'pentaho.viz.crossfilter',          // type of the Javascript object to instantiate
    needsColorGradient: false,
    customType: '',
    maxValues: [50000,50000,50000,50000],
    getDropZoneLabel: function(type){
      var label = type;
      return label;
    },
    args: {
    },
    propMap: [
    ],
    helper: {
        canRefreshReport: function(report) {
          var ok = true;
          if (report.findGemsByGembarId("rows").length == 0 || report.findGemsByGembarId("measures").length == 0)
            ok = false;
          return ok;
        }
    },
    dataReqs: [                             // dataReqs describes the data requirements of this visualization
        {
            name: 'Default',
            reqs: 
                [
                    {   id: 'rows',             // id of the data element
                        dataType: 'string',         // data type - 'string', 'number', 'date', 'boolean', 'any' or a comma separated list
                        dataStructure: 'column',    // 'column' or 'row' - only 'column' supported so far
                        caption: 'rows',        // visible name
                        required: true,              // true or false
                        allowMultiple: true         // true or false
                    },
                    {   id: 'measures',
                        dataType: 'number',
                        dataStructure: 'column',
                        caption: 'measures',
                        required: true,
                        allowMultiple: true         // true or false
                    }
                ]
        }
    ]
});

/*
    CccChart constructor
    This takes an HTML DOM element as a parameter
*/
pentaho.viz.crossfilter = function( element ) {
    this.debug = false;
    this.element = element;
    this.elementName = element.id;
    this.cdaTable = null;
    this.dataTable = null;
    this.vizOptions = null;
    this.series = [];
    this.selections = [];
    this.dimensionFunctions = [];
    this.crossf = null; // the crossfilter
    this.dimensions = [];
    this.dimGroups = [];
    this.ranges = [];
    this.charts = [];
    this.sortIdx = 0;
    crossFilterWrapper = this;
    
}

pentaho.viz.crossfilter.prototype.resize = function() {
    this.draw(this.dataTable, this.vizOptions);
}

/*
    draw()
    
    dataTable   a pentaho.DataTable object with the data to display
    vizOptions  the options for the visualization
*/
pentaho.viz.crossfilter.prototype.draw = function( pdataTable, vizOptions ) {

    if( this.debug ) console.log('pentaho.viz.crossfilter.draw()');
    
    // detect IE
    var browserName=navigator.appName;
    if (browserName=="Microsoft Internet Explorer") {
        this.element.innerHTML = pentaho.common.Messages.getString('noIE');
        return;
    }
    
    this.dimensions = [];
    dimensions = [];
    this.dimGroups = [];
    this.ranges = [];
    this.charts = [];
    this.sortIdx = 0;
    this.crossf = null; // the crossfilter

    this.dimensionFunctions = [];
    this.dimensionFunctions[0] = function(d) { return d["measure0"] };
    this.dimensionFunctions[1] = function(d) { return d["measure1"] };
    this.dimensionFunctions[2] = function(d) { return d["measure2"] };
    this.dimensionFunctions[3] = function(d) { return d["measure3"] };
    this.dimensionFunctions[4] = function(d) { return d["measure4"] };
    this.dimensionFunctions[5] = function(d) { return d["measure5"] };
    this.dimensionFunctions[6] = function(d) { return d["measure6"] };
    this.dimensionFunctions[7] = function(d) { return d["measure7"] };

    // store the incoming parameters for later use
    this.vizOptions = vizOptions;
    this.dataTable = pdataTable;
    dataTable = pdataTable;
    
    // store the current highlighted selections
    this.selections = vizOptions.selections;

    // local variables
    var metadata = [];
    var measures = [];
    var strings = [];
    var resultset = [];
    var colors = null;
    var columnMap = {};
    
    var title = '';
    measureNo = 0;
    for( var colNo=0; colNo<dataTable.getNumberOfColumns(); colNo++) {
        if( dataTable.getColumnType(colNo).toUpperCase() == 'NUMBER' ) {
            measures.push(colNo);
            title = dataTable.getColumnLabel(colNo);
            columnMap[ colNo ] = [title,'measure'+measureNo];
            measureNo++;
        }
        else {
            title = dataTable.getColumnLabel(colNo);
            columnMap[ colNo ] = [title];
            strings.push(colNo);
        }
    }

    var rows = [];

    for( var rowNo=0; rowNo<dataTable.getNumberOfRows(); rowNo++ ) {
        var row = {};
        for( var colNo=0; colNo<dataTable.getNumberOfColumns(); colNo++ ) {
            var mapping = columnMap[colNo];
            for( mapNo=0; mapNo<mapping.length; mapNo++) {
                row[ mapping[mapNo] ] = dataTable.getValue(rowNo, colNo); 
                row[ 'formatted_'+mapping[mapNo] ] = dataTable.getFormattedValue(rowNo, colNo); 
            }
        }
        rows.push( row );
    }

    if( this.crossf ) {
//        this.crossf.reset();
        this.crossf = new crossfilter( rows );

    } else {
        this.crossf = new crossfilter( rows );
    }
    // create dimensions
    
    for( var idx=0; idx<measures.length; idx++ ) {
        var measureId = dataTable.getColumnId(measures[idx]);

        var func = this.dimensionFunctions[idx];
        var dimension = this.crossf.dimension(func);
        this.dimensions.push( dimension );
        dimensions.push( dimension );
        this.ranges.push( dataTable.getColumnRange(measures[idx]) );
    }

    all = this.crossf.groupAll();
    for( var idx=0; idx<this.dimensions.length; idx++ ) {
        var range = this.ranges[idx].max - this.ranges[idx].min;
        var bucket = range/100;
        // adjust bucket to nearest 1,2,5,10,20,50 etc
        var group = this.dimensions[idx].group(function(value) { return Math.floor(value / bucket)*bucket; });
        this.dimGroups.push( group );
    }

    while(this.element.firstChild) {
        this.element.removeChild(this.element.firstChild);
    }
    
    var chartsPerRow = Math.floor(this.getClientWidth() / 350);
        
    var table = document.createElement('TABLE');
    table.id = 'charttable';
    this.element.appendChild(table);
    var row;
    
    for(idx=0; idx<this.dimensions.length; idx++) {
        if( idx % chartsPerRow == 0 ) {
            row = table.insertRow(-1);
        }
        var cell = row.insertCell(-1);
        var span = document.createElement('SPAN');
        span.className = 'chartTitle';
        span.innerHTML = dataTable.getColumnLabel(measures[idx]);
        cell.appendChild(span);
        var span = document.createElement('SPAN');
        span.className = 'chartOp';
        span.innerHTML = ' <span onclick="crossFilterWrapper.reset ('+idx+')">[show all]</span>';
        cell.appendChild(span);
        var br = document.createElement('BR');
        cell.appendChild(br);
        var div = document.createElement('DIV');
        div.className = 'chart';
        cell.appendChild(div);
    }
    
    var table = document.createElement('TABLE');
    table.id = 'datalist';
    table.className = 'datatable';
    table.cellPadding = "0";
    table.cellSpacing = "0";
    
    this.element.appendChild(table);
    var row = table.insertRow(-1);
    var cell = row.insertCell(0);
    cell.innerHTML = "#";  
    cell.className = 'tableheader alignRight';
    var measureNo = 0;
    for( var colNo=0; colNo<dataTable.getNumberOfColumns(); colNo++) {
        var cell = row.insertCell(colNo+1);
        var sortable = '';
        if( dataTable.getColumnType(colNo).toUpperCase() == 'NUMBER' ) {
            cell.className = 'tableheader alignRight';
            sortable = ' <span onClick="crossFilterWrapper.setSort('+measureNo+')"><></span>';
            measureNo++;
        } else {
            cell.className = 'tableheader';
        }
        cell.innerHTML = dataTable.getColumnLabel(colNo) + sortable;  
    }

    formatNumber = d3.format(",d");

    this.charts = [];

    for( var idx=0; idx<this.dimensions.length; idx++ ) {
        if( idx < this.charts.length ) {
            this.charts[idx].reset();
            this.charts[idx].dimension(this.dimensions[idx])
                .group(this.dimGroups[idx])
                .x(d3.scale.linear()
                .domain([this.ranges[idx].min*0.9, this.ranges[idx].max*1.1])
                .rangeRound([0, 300]));
        } else {
            this.charts[idx] = this.barChart(idx)
                .dimension(this.dimensions[idx])
                .group(this.dimGroups[idx])
                .x(d3.scale.linear()
                .domain([this.ranges[idx].min*0.9, this.ranges[idx].max*1.1])
                .rangeRound([0, 300]));
        }
    }       

    chart = d3.selectAll(".chart")
        .data(this.charts)
        .each(function(chart) { chart.on("brush", crossFilterWrapper.renderAll).on("brushend", crossFilterWrapper.renderAll); });    
    
    list = d3.selectAll(".datatable")
      .data([this.updateTable]);
          
  this.renderAll();

    if( this.debug ) console.log('pentaho.viz.crossfilter.draw() done');
        
}

pentaho.viz.crossfilter.prototype.getClientWidth = function() {
      return this.element.offsetWidth;
};


pentaho.viz.crossfilter.prototype.barChart = function(id) {

    var margin = {top: 10, right: 10, bottom: 20, left: 10},
        x,
        y = d3.scale.linear().range([100, 0]),
        axis = d3.svg.axis().orient("bottom"),
        brush = d3.svg.brush(),
        brushDirty,
        dimension,
        group,
        round;

    function reset() {
        x = null;
        y = d3.scale.linear().range([100, 0]);
        axis = d3.svg.axis().orient("bottom");
        brush = d3.svg.brush();
        brushDirty = false;
        dimension = null;
        group = null;
        round = null;
    };

    function chart(div) {
      var width = x.range()[1],
          height = y.range()[0];

      y.domain([0, group.top(1)[0].value]);

      div.each(function() {
        var div = d3.select(this),
            g = div.select("g");

        // Create the skeletal chart.
        if (g.empty()) {
          div.select(".title").append("a")
              .attr("href", "javascript:reset(" + id + ")")
              .attr("class", "reset")
              .text("reset")
              .style("display", "none");

          g = div.append("svg")
              .attr("width", width + margin.left + margin.right)
              .attr("height", height + margin.top + margin.bottom)
            .append("g")
              .attr("transform", "translate(" + margin.left + "," + margin.top + ")");

          g.append("clipPath")
              .attr("id", "clip-" + id)
            .append("rect")
              .attr("width", width)
              .attr("height", height);

          g.selectAll(".bar")
              .data(["background", "foreground"])
            .enter().append("path")
              .attr("class", function(d) { return d + " bar"; })
              .datum(group.all());

          g.selectAll(".foreground.bar")
              .attr("clip-path", "url(#clip-" + id + ")");

          g.append("g")
              .attr("class", "axis")
              .attr("transform", "translate(0," + height + ")")
              .call(axis);

          // Initialize the brush component with pretty resize handles.
          var gBrush = g.append("g").attr("class", "brush").call(brush);
          gBrush.selectAll("rect").attr("height", height);
          gBrush.selectAll(".resize").append("path").attr("d", resizePath);
        }

        // Only redraw the brush if set externally.
        if (brushDirty) {
          brushDirty = false;
          g.selectAll(".brush").call(brush);
          div.select(".title a").style("display", brush.empty() ? "none" : null);
          if (brush.empty()) {
            g.selectAll("#clip-" + id + " rect")
                .attr("x", 0)
                .attr("width", width);
          } else {
            var extent = brush.extent();
            g.selectAll("#clip-" + id + " rect")
                .attr("x", x(extent[0]))
                .attr("width", x(extent[1]) - x(extent[0]));
          }
        }

        g.selectAll(".bar").attr("d", barPath);
      });

      function barPath(groups) {
        var path = [],
            i = -1,
            n = groups.length,
            d;
        while (++i < n) {
          d = groups[i];
          path.push("M", x(d.key), ",", height, "V", y(d.value), "h9V", height);
        }
        return path.join("");
      }

      function resizePath(d) {
        var e = +(d == "e"),
            x = e ? 1 : -1,
            y = height / 3;
        return "M" + (.5 * x) + "," + y
            + "A6,6 0 0 " + e + " " + (6.5 * x) + "," + (y + 6)
            + "V" + (2 * y - 6)
            + "A6,6 0 0 " + e + " " + (.5 * x) + "," + (2 * y)
            + "Z"
            + "M" + (2.5 * x) + "," + (y + 8)
            + "V" + (2 * y - 8)
            + "M" + (4.5 * x) + "," + (y + 8)
            + "V" + (2 * y - 8);
      }
    }

    brush.on("brushstart.chart", function() {
      var div = d3.select(this.parentNode.parentNode.parentNode);
      div.select(".title a").style("display", null);
    });

    brush.on("brush.chart", function() {
      var g = d3.select(this.parentNode),
          extent = brush.extent();
      if (round) g.select(".brush")
          .call(brush.extent(extent = extent.map(round)))
        .selectAll(".resize")
          .style("display", null);
      g.select("#clip-" + id + " rect")
          .attr("x", x(extent[0]))
          .attr("width", x(extent[1]) - x(extent[0]));
      dimension.filterRange(extent);
    });

    brush.on("brushend.chart", function() {
      if (brush.empty()) {
        var div = d3.select(this.parentNode.parentNode.parentNode);
        div.select(".title a").style("display", "none");
        div.select("#clip-" + id + " rect").attr("x", null).attr("width", "100%");
        dimension.filterAll();
      }
    });

    chart.margin = function(_) {
      if (!arguments.length) return margin;
      margin = _;
      return chart;
    };

    chart.x = function(_) {
      if (!arguments.length) return x;
      x = _;
      axis.scale(x);
      brush.x(x);
      return chart;
    };

    chart.y = function(_) {
      if (!arguments.length) return y;
      y = _;
      return chart;
    };

    chart.dimension = function(_) {
      if (!arguments.length) return dimension;
      dimension = _;
      return chart;
    };

    chart.filter = function(_) {
      if (_) {
        brush.extent(_);
        dimension.filterRange(_);
      } else {
        brush.clear();
        dimension.filterAll();
      }
      brushDirty = true;
      return chart;
    };

    chart.group = function(_) {
      if (!arguments.length) return group;
      group = _;
      return chart;
    };

    chart.round = function(_) {
      if (!arguments.length) return round;
      round = _;
      return chart;
    };

    return d3.rebind(chart, brush, "on");
  }

// Renders the specified chart or list.
  function render(method) {
    if( this.debug ) console.log('pentaho.viz.crossfilter.render()');
    d3.select(this).call(method);
    if( this.debug ) console.log('pentaho.viz.crossfilter.render() done');
  }

  // Whenever the brush moves, re-rendering everything.
  pentaho.viz.crossfilter.prototype.renderAll = function() {
    if( this.debug ) console.log('pentaho.viz.crossfilter.renderAll()');
    chart.each(render);
    list.each(render);
    d3.select("#active").text(formatNumber(all.value()));
    if( this.debug ) console.log('pentaho.viz.crossfilter.renderAll() done');
  }

  window.filter = function(filters) {
    if( this.debug ) console.log('pentaho.viz.crossfilter.filter()');
    filters.forEach(function(d, i) { charts[i].filter(d); });
    crossFilterWrapper.renderAll();
    if( this.debug ) console.log('pentaho.viz.crossfilter.filter() done');
  };

  pentaho.viz.crossfilter.prototype.reset = function(i) {
    if( this.debug ) console.log('pentaho.viz.crossfilter.reset()');
    this.charts[i].filter(null);
    this.renderAll();
    if( this.debug ) console.log('pentaho.viz.crossfilter.reset() done');
  };
  
  pentaho.viz.crossfilter.prototype.setSort = function( idx ) {
    if( this.debug ) console.log('pentaho.viz.crossfilter.setSort()');
    this.sortIdx = idx;
    this.renderAll();
    if( this.debug ) console.log('pentaho.viz.crossfilter.setSort() done');
  }
  
  pentaho.viz.crossfilter.prototype.updateTable = function(div) {
    if( this.debug ) console.log('pentaho.viz.crossfilter.updateTable()');

    var data = dimensions[crossFilterWrapper.sortIdx].top(50);

    var table = document.getElementById('datalist');
    while( table.rows.length>1) {
        table.deleteRow(1);
    }

    var colCount = dataTable.getNumberOfColumns();
    var rowCount = data.length;
    for( var idx=0; idx<rowCount; idx++ ) {
        var row = table.insertRow(idx+1);
        var cell = row.insertCell(0);
        cell.innerHTML = ''+(idx+1);
        var rowClass = idx % 2 ? ' oddRow' : '';
        var className = 'tableCell' + rowClass;
        if( idx == rowCount-1 ) {
            className += ' lastRow';
        }
        cell.className = className+' alignRight';
        for( var colNo=0; colNo<colCount; colNo++) {
            var cell = row.insertCell(colNo+1);
            var value = data[idx]['formatted_'+dataTable.getColumnLabel(colNo)];
            if(value == undefined) {
                value = data[idx][dataTable.getColumnLabel(colNo)];
            }
            cell.innerHTML = value;  
            if( dataTable.getColumnType(colNo).toUpperCase() == 'NUMBER' ) {
                className += ' alignRight';
            }
            if( colNo == colCount-1 ) {
                className += ' lastColumn';
            }
            cell.className = className;
        }
    }
    
    if( this.debug ) console.log('pentaho.viz.crossfilter.updateTable() done');
    return;
  
  }
